
read_ordering( Ord ):-
        read( Ord1 ),
	check_and_transform_ordering( Ord1, Ord ),!.


check_and_transform_ordering( mspo(int(I,B),Q), mspo(int(I,B1),Q1)  ):-
	check_and_transform_base( B, B1 ),
	check_interpretation( I ),
	check_and_transform_q( int(I,B1), Q, Q1 ),!.



check_and_transform_base( rpo(Prec,Stat), rpo(Prec1,Stat) ):-
	check_and_close(Prec,Prec1),!.
check_and_transform_base(_,_):-
	write('Incorrect Base Ordering!'),nl,nl,!.

check_and_close(Prec,Prec1):-
	transitive_closure(Prec,Prec1),
	\+ contradictory(Prec1),!.

contradictory(Prec):-
	member22(X>Y,Prec), 
	( member212(Y>X,Prec) ; member212(X=Y,Prec)),!.
contradictory(Prec):-
	member22(X=Y,Prec), 
	( member212(Y>X,Prec) ; member212(X>Y,Prec)),!.


transitive_closure(Prec, Prec1):-
	list_operators(Prec,L0),
	sort(L0,L),
	cartesian_product(L,L,[],Pairs),
	tclose(Pairs,Prec,Prec2),
	sort(Prec2,Prec1),!.


list_operators([],[]).
list_operators([F>G|Ops],[F,G|L]):-
	list_operators(Ops,L).
list_operators([F=G|Ops],[F,G|L]):-
	list_operators(Ops,L).



cartesian_product([],_,P,P).
cartesian_product([X|L1],L2,P1,P2):-
	cart(X,L2,P1,P3),
	cartesian_product(L1,L2,P3,P2).

cart(_,[],P,P).
cart(X,[Y|L],P1,[[X,Y]|P2]):-
	cart(X,L,P1,P2).


tclose([],_,[]).
tclose([[X,X]|L],Prec,Prec1):- !,
	tclose(L,Prec,Prec1),!.
tclose([[X,Y]|L],Prec,[X>Y|Prec1]):-
	gr_in_closure(X,Y,Prec),!,
	tclose(L,Prec,Prec1),!.
tclose([[X,Y]|L],Prec,[F=G|Prec1]):-
	eq_in_closure(X,Y,Prec),
	sort([X,Y],[F,G]),!,
	tclose(L,Prec,Prec1),!.
tclose([_|L],Prec,Prec1):-
	tclose(L,Prec,Prec1),!.


gr_in_closure(Op1, Op2, Pre):- member22(Op1>Op2, Pre),!.
gr_in_closure(Op1, Op2, Pre):-
	delete32(Op1>Op3,Pre,Pre1),
	gr_in_closure(Op3,Op2, Pre1),!.
gr_in_closure(Op1, Op2, Pre):-
	delete32(Op1>Op3,Pre,Pre1),
	eq_in_closure(Op3,Op2, Pre1),!.
gr_in_closure(Op1, Op2, Pre):-
	delete32(Op1=Op3,Pre,Pre1),
	gr_in_closure(Op3,Op2, Pre1),!.
gr_in_closure(Op1, Op2, Pre):-
	delete32(Op3=Op1,Pre,Pre1),
	gr_in_closure(Op3,Op2, Pre1),!.

eq_in_closure(Op1, Op2, Pre):- member22(Op1=Op2, Pre),!.
eq_in_closure(Op1, Op2, Pre):- member22(Op2=Op1, Pre),!.
eq_in_closure(Op1, Op2, Pre):-
	delete32(Op1=Op3,Pre,Pre1),
	eq_in_closure(Op3,Op2, Pre1),!.
eq_in_closure(Op1, Op2, Pre):-
	delete32(Op3=Op1,Pre,Pre1),
	eq_in_closure(Op3,Op2, Pre1),!.



%------------------------------------------------------------------------------
% check_interpretation(Int):
%
% 1) [f(x,y,z), x]	: I(f(t1,t2,t3))=I(t1)
% 2) [f(x,y,z), g(x,z)]	: I(f(t1,t2,t3))=g(I(t1),I(t3))
%
% There can be no two different interpretations for the same f
%
%------------------------------------------------------------------------------

check_interpretation([]):- !.
check_interpretation([[L,R]|Rest] ):-
	check_int(L,R),
	nonvar(L),
	functor(L,F,_),
	\+ another_interpretation(F,Rest),
	check_interpretation(Rest), !.
check_interpretation([I|_] ):-
	nl, nl, write('Wrong definition: '), write(I), 
	write(' of the interpretation'), nl, !, fail.

another_interpretation(F,I):-
	member22([L,_],I),
	functor(L,F,_),!,fail.

check_int(L,R):-
	nonvar(L),
	var(R),
	L =..[_|Args],
	all_distinct_vars(Args),
	member212(R, Args), !.
check_int(L,R):-
	nonvar(L), nonvar(R),
	L =..[_|ArgsL],
	R =..[_|ArgsR],
	all_distinct_vars(ArgsL),
	all_are_args_of_L( ArgsR, ArgsL ),!.


all_distinct_vars([]).
all_distinct_vars([X|Args]):-
	member212(X,Args), !, fail.
all_distinct_vars([X|Args]):-
	var(X),	
	all_distinct_vars(Args),!.


all_are_args_of_L( [], _ ):- !.
all_are_args_of_L( [S|R], ArgsL):-
	member212( S, ArgsL ),
	all_are_args_of_L( R, ArgsL ), !.


%------------------------------------------------------------------------------
% check_and_transform_q( I, Q, Q1 )
%------------------------------------------------------------------------------

check_and_transform_q( I, int, I ):-!.
check_and_transform_q( _, pre(P), pre(P1) ):-
	check_and_close(P,P1),!.	
check_and_transform_q( I, ren(R,Q), ren(R,Q1) ):-
	check_renaming(R),
	check_and_transform_q( I, Q, Q1 ),!.
check_and_transform_q( I, Lex, Lex1 ):-
	Lex  =.. [lex|Comps],
	check_and_transform_qs( I, Comps, Comps1 ),
	Lex1 =.. [lex|Comps1],!.
check_and_transform_q( _,_,_ ):- write('Incorrect Q'),nl,nl,fail.


check_and_transform_qs( _, [], [] ).
check_and_transform_qs( I, [Q|Comps], [Q1|Comps1] ):-
	check_and_transform_q( I, Q, Q1 ),
	check_and_transform_qs( I, Comps, Comps1 ),!.

check_renaming([]):- !.
check_renaming([ren(F)=_|R]):-
	member22(ren(F)=_, R), !,
	lwrite([nl,'Wrong renaming: symbol ',F,' renamed twice!',nl]), !,fail.
check_renaming([ren(_)=_|R]):- check_renaming(R),!.


%------------------------------------------------------------------------------
% display_ordering(Order)
%------------------------------------------------------------------------------

display_ordering( mspo(int(Int,Base), Q) ):-
	nl, nl, 
	lwrite(['MSPO with:', nl ]),
	lwrite(['  Base ordering:  ', nl, '    ', Base, nl ]),
	lwrite(['  Interpretation: ', nl ]), display_interpretation(Int),
	lwrite(['  Component Q:    ', nl ]), display_q(Q,4),!.


display_q( int(_,_), N ):-
	tab(N),
	write(int),nl,!.
display_q( pre(P), N ):-
	tab(N),
	write(pre(P)), nl,!.
display_q( ren(R,Q), N ):-
	tab(N),
	write('ren('),nl,
	N1 is N+4,
	tab(N1), write(R), nl,
	display_q(Q,N1),
	tab(N), write(')'),nl,!.
display_q( L, N ):-
	L =.. [lex|Comps],
	tab(N),
	write('lex( '), nl,
	N1 is N+4,
	display_qs(Comps,N1),
	tab(N), write(')'),nl,!.


display_qs([],_):-!.
display_qs([Q|R],N):-
	display_q(Q,N),
	display_qs(R,N),!.


display_interpretation([]):- !.
display_interpretation([I|Rest]):-
	lwrite(['    ', I, nl]),
	display_interpretation(Rest), !.	
	
% eof-------------------------------------------------------------------


